"""Estimate Functions Using Local Polynomials."""
import math
from typing import Optional, Union

from mypy_extensions import TypedDict  # Python>=3.8: from typing import TypedDict
from numba import njit
import numpy as np
import pandas as pd

from kernreg.bandwidth_selection import get_bandwidth_rsc
from kernreg.funcs_to_jit import (
    combine_bincounts_kernelweights,
    get_kernelweights,
    is_sorted,
)
from kernreg.linear_binning import linear_binning
from kernreg.utils import process_inputs


# Jit the functions
# Implemented as additional step since pytest-cov does not consider
# jitted functions when determining code coverage.
# The respective un-jitted functions are hence unit-tested in tests/test_locpoly.py
linear_binning_jitted = njit(linear_binning)
is_sorted_jitted = njit(is_sorted)
get_weights_jitted = njit(get_kernelweights)
combine_bincounts_weights_jitted = njit(combine_bincounts_kernelweights)


class Result(TypedDict):
    """Result dict for func ``locpoly``."""

    gridpoints: np.ndarray
    curvest: np.ndarray
    bandwidth: float


def locpoly(
    x: Union[np.ndarray, pd.Series],
    y: Union[np.ndarray, pd.Series],
    derivative: int = 0,
    degree: Optional[int] = None,
    gridsize: Optional[int] = None,
    bandwidth: Optional[float] = None,
    a: Optional[float] = None,
    b: Optional[float] = None,
    binned: bool = False,
    truncate: bool = True,
) -> Result:
    r"""Estimates a regression function or their derivatives using local polynomials.

    Non-parametrically fits a smooth curve between a predictor, ``x``,
    and a response variable, ``y``. A binned approximation to ``x`` and ``y``
    over an equally-spaced grid is used for fast computation.

    Note that for a `v`-th derivative the order of the polynomial
    must be :math:`p = v + 1`, :math:`v + 3`, :math:`v + 5`, or :math:`v + 7`.

    The local polynomial curve estimator ``beta`` and its derivatives are
    minimizers to the locally weighted least-squares problem.
    In particular, at each point :math:`g = 1,..., M` in the grid,
    ``beta`` is computed as the solution to the linear matrix equation:

    .. math::

        X'W X \ \beta = X'W y,

    where :math:`W` are kernel weights approximated by the Gaussian density function.
    :math:`X'W X` and :math:`X'W y` use binned approximations to ``X`` and ``y``,
    respectively; see :func:`~kernreg.linear_binning.linear_binning`.

    The terms "kernel" and "kernel function" are used interchangeably
    throughout.

    :cite:`Fan1994` recommend a default of ``gridsize`` = 400
    for the popular case of graphical analysis. They find that using fewer than 400
    grid points produces distracting "granularity", while more grid points
    often give negligible improvements in resolution.

    The user may pre-specifiy a bandwidth smoothing parameter.
    If not, the bandwidth is auto-selected. ``locpoly`` uses the bandwidth
    that minimzes the Residual Squares Criterion (RSC), see
    :func:`~kernreg.bandwidth_selection.get_bandwidth_rsc`.

    Arguments:
        x: Array of x data. Missing values are not accepted. Must be sorted.
        y: Array of y data. This must be same length as x. Missing values are
            not accepted. Must be presorted by x.
        derivative: Order of the derivative to be estimated.
            ``derivative`` = 0 means a function, i.e. `0`-th derivative, is fitted.
        degree: Degree of local polynomial used. Its value must be equal to
            ``derivative + 1`` or ``derivative + 3``.
        bandwidth: Kernel bandwidth smoothing parameter. If not specified,
            the bandwidth that minimizes the RSC is chosen.
        gridsize: Number of equally-spaced grid points over which the
            function is to be estimated.
        a: Start point of the grid.
        b: End point of the grid.
        binned: If True, then ``x`` and ``y`` are taken to be bin counts
            rather than raw data and the binning step is skipped.
        truncate: If True, trim endpoints.

    Returns:
        Result: Result Dictionary containing:
            - gridpoints (np.ndarray): Sorted grid points (``x-dimension``) at
                which the estimate of :math:`E[Y|X]` (or its derivative) is computed
            - curvest (np.ndarray): Curve estimate for the specified
                derivative of ``beta``
            - bandwidth (float): Bandwidth used. If not pre-specified by
                the user, this is the bandwidth that minimizes the RSC.

    Raises:
        Exception: Input data ``x`` and ``y`` must be sorted by ``x``
            before estimation.
        Exception: The degree of the polynomial must be equal to
            ``derivative + 1``, ``derivative + 3``, ``derivative + 5``, or
             ``derivative + 7``.

    """
    x, y, degree, gridsize, a, b = process_inputs(
        x, y, derivative, degree, gridsize, a, b
    )

    if is_sorted_jitted(x) is False:
        raise Exception("Input arrays x and y must be sorted by x before estimation!")

    if degree not in [derivative + 1, derivative + 3, derivative + 5, derivative + 7]:
        raise Exception(
            "The degree of the polynomial must be equal to derivative "
            "v + 1, v + 3, v + 5, or v + 7."
        )

    if bandwidth is None:
        bandwidth = get_bandwidth_rsc(x, y, derivative, degree, [a, b])

    # Set bin width
    binwidth = (b - a) / (gridsize - 1)

    # 1. Bin the data if not already binned
    if binned is False:
        xcounts, ycounts = linear_binning_jitted(x, y, gridsize, a, binwidth, truncate)
    else:
        xcounts, ycounts = x, y

    # 2. Obtain kernel weights
    # Note that only L < N kernel evaluations are required to obtain the
    # kernel weights regardless of the number of observations N.
    # get_weights_jitted = njit(get_kernelweights)
    weights = get_weights_jitted(bandwidth, binwidth)

    # 3. Combine bin counts and kernel weights
    # combine_bincounts_weights_jitted = njit(combine_bincounts_weights)
    x_weighted, y_weighted = combine_bincounts_weights_jitted(
        xcounts, ycounts, weights, degree, gridsize, bandwidth, binwidth
    )

    # 4. Fit the curve and obtain estimator for the desired derivative
    curvest = get_curve_estimator(x_weighted, y_weighted, degree, derivative, gridsize)

    # Generate grid points for visual representation
    gridpoints = np.linspace(a, b, gridsize)

    rslt = Result(gridpoints=gridpoints, curvest=curvest, bandwidth=bandwidth)

    return rslt


def get_curve_estimator(
    x_weighted: np.ndarray,
    y_weighted: np.ndarray,
    degree: int,
    derivative: int,
    gridsize: int,
) -> np.ndarray:
    r"""Solves the locally weighted least-squares regression problem.

    Returns an estimator for the `v`-th derivative of ``beta`` .

    The local polynomial curve estimator ``beta`` and its derivatives are
    minimizers to a locally weighted least-squares problem. At each grid
    point, ``beta`` is computed as the solution to the linear matrix equation:

    .. math::

        X'W X \ \beta = X'W y,

    where W are kernel weights approximated by the Gaussian density function.
    :math:`X'W X`and :math:`X'W y`are approximated by ``x_weighted`` and ``y_weighted``,
    which are the result of a direct convolution of
    bin counts (``xcounts``) and kernel weights, as well as
    bin averages (``ycounts``) and kernel weights, respectively.

    Arguments:
        x_weighted: Binned approximation to :math:`X'W X`.
        y_weighted: Binned approximation to :math:`X'W y`.
        degree: Degree of the polynomial.
        derivative: Order of the derivative to be estimated.
        gridsize: Number of equally-spaced grid points.

    Returns:
        curvest: Curve estimate for the specified derivative of ``beta``.

    """
    coly = degree + 1
    xmat = np.zeros((coly, coly))
    yvec = np.zeros(coly)
    curvest = np.zeros(gridsize)

    for g in range(gridsize):
        for row in range(0, coly):
            for column in range(0, coly):
                colindex = row + column
                xmat[row, column] = x_weighted[g, colindex]

                yvec[row] = y_weighted[g, row]

        # Calculate beta as the solution to the linear matrix equation
        # X'W X * beta = X'W y.
        # Note that xmat and yvec are binned approximations to X'W X and
        # X'W y, respectively, evaluated at the given grid point.
        beta = np.linalg.solve(xmat, yvec)

        # Obtain curve estimator for the desired derivative of beta.
        curvest[g] = beta[derivative]

    curvest = math.gamma(derivative + 1) * curvest

    return curvest



"""General configuration for kernreg."""
from pathlib import Path

# Obtain the root directory of the package.
# Do not import kernreg which creates a circular import.
ROOT_DIR = Path(__file__).parent.parent.parent

# Directory with additional resources for the testing harness
TEST_DIR = ROOT_DIR / "tests"
TEST_RESOURCES_DIR = ROOT_DIR / "tests" / "resources"
RESOURCES_DIR = ROOT_DIR / "src" / "kernreg" / "example_data"



"""Bandwidth selection via the Residual Squares Criterion (RSC)."""
from functools import partial
import math
from typing import List, Optional, Union

import numpy as np


def get_bandwidth_rsc(
    x: np.ndarray,
    y: np.ndarray,
    derivative: int,
    degree: int,
    x_range: Optional[Union[List[float], np.ndarray]] = None,
) -> float:
    r"""Returns the bandwidth that minimizes the Residual Squares Criterion.

    The Residual Squares Criterion (RSC) is definied as (:cite:`Fan1996`, p. 118):

    .. math::

        \text{RSC}(x_0; h) = \hat{\sigma}^2(x_0) \{ 1 + (p + 1) V \},

    where `V` is the first diagonal element of the matrix
    :math:`(\textbf{X}^T \textbf{W} \textbf{X})^{-1}
    (\textbf{X}^T \textbf{W}^2 \textbf{X})
    (\textbf{X}^T \textbf{W} \textbf{X})^{-1}`.

    The expression :math:`(\textbf{X}^T \textbf{W} \textbf{X})^{-1}
    (\textbf{X}^T \textbf{W}^2 \textbf{X})
    (\textbf{X}^T \textbf{W} \textbf{X})^{-1} \sigma^2`
    is an approximation to the conditional variance
    :math:`\text{Var}(\hat{\beta} \,|\, \textbf{X})`.

    Weight matrix `W` contains the diagonal elements of :math:`K_n(X_i - x_0)`,
    where kernel function :math:`K_n` represents the Gaussian density.

    The unkown quantity :math:`\hat{\sigma}^2`
    can be estimated by the normalized weighted residual sum of squares,
    see :func:`~kernreg.bandwidth_selection.get_residual_squares_criterion`:

    .. math::
        \hat{\sigma}^2(x_0) = \frac{\sum\limits_{i=1}^n
        (Y_i - \hat{Y}_i)^2 K_{h^*} (X_i - x_0)}
        { \text{tr} \{ \textbf{W}^* - \textbf{W}^* \textbf{X}^*
        ( \textbf{X}^{*T} \textbf{W}^* \textbf{X}^*)^{-1}
        \textbf{X}^{*T} \textbf{W}^*\} }.

    The RSC is computed for various combinations of ``x_0`` and bandwidth ``bw``.
    The RSCs for a particular bandwidth are then summed together across
    the ``x_0``. Finally, the bandwidth with the smallest sum of RCSs
    is chosen.


    Arguments:
        x: Predictor variable.
        y: Response variable.
        derivative: Order of the derivative to be estimated.
        degree: Degree of the polynomial.
        x_range: Start and end point of the domain of ``x`` to consider.
            Default is the entire range of ``x``.

    Returns:
        Bandwidth that minimizes the residual squares criterion.
            Adjusted by constant for the Gaussian kernel.
            See :cite:`Fan1996`, p. 120, and :cite:`Fan1995`.

    """
    if x_range is not None:
        xmin = x_range[0]
        xmax = x_range[1]
    else:
        xmin = min(x)
        xmax = max(x)

    # Construct sequence of 11 x-values to evaluate bandwidth on
    n_evals = 11
    xseq = np.linspace(xmin, xmax, n_evals)

    bw_min = (xmax - xmin) / 20
    bw_max = (xmax - xmin) / 2
    bw_cons = 1.1

    # Determine total number of bandwidths to iterate over
    total = math.ceil(np.log(bw_max / bw_min) / np.log(bw_cons))
    bw_arr = np.array([bw_min * (bw_cons ** (i + 1)) for i in range(total)])
    cartesian_pairs = np.transpose(
        [np.tile(xseq, len(bw_arr)), np.repeat(bw_arr, n_evals)]
    )

    rsc_p = partial(get_residual_squares_criterion, x, y, degree)

    sums_1d = np.zeros(total * n_evals)
    for index, value in enumerate(cartesian_pairs):
        rsc = rsc_p(value)
        sums_1d[index] += rsc

    rsc_sums = np.sum(sums_1d.reshape(total, n_evals), axis=1)

    # Adjusting constants for the Gaussian kernel, see :cite:`Fan1995`.
    if degree == derivative + 1:
        const = np.array([1, 0.8403, 0.8285, 0.8085, 0.8146, 0.8098, 0.8159])
    elif degree == derivative + 3:
        const = [np.nan] * 2 + [0.9554, 0.8975, 0.8846, 0.8671, 0.8652]
    elif degree == derivative + 5:
        const = [np.nan] * 4 + [0.9495, 0.9165, 0.9055]
    else:  # degree = derivative + 7
        const = [np.nan] * 6 + [0.9470]

    bw = float(bw_arr[np.where(rsc_sums == np.min(rsc_sums))])
    bw_adj = bw * const[degree - 1]

    return bw_adj


def get_residual_squares_criterion(
    x: np.ndarray, y: np.ndarray, degree: int, input_arr: Union[List[float], np.ndarray]
) -> float:
    r"""Compute residual squares criterion for given ``x_0`` and ``bw``.

    The Residual Squares Criterion (RSC) is definied as (:cite:`Fan1996`, p. 118):

    .. math::

        \text{RSC}(x_0; h) = \hat{\sigma}^2(x_0) \{ 1 + (p + 1) V \},

    where `V` is the first diagonal element of the matrix
    :math:`(\textbf{X}^T \textbf{W} \textbf{X})^{-1}
    (\textbf{X}^T \textbf{W}^2 \textbf{X})
    (\textbf{X}^T \textbf{W} \textbf{X})^{-1}`.

    The expression :math:`(\textbf{X}^T \textbf{W} \textbf{X})^{-1}
    (\textbf{X}^T \textbf{W}^2 \textbf{X})
    (\textbf{X}^T \textbf{W} \textbf{X})^{-1} \sigma^2`
    is an approximation to the conditional variance
    :math:`\text{Var}(\hat{\beta} \,|\, \textbf{X})`.

    Weight matrix `W` contains the diagonal elements of :math:`K_n(X_i - x_0)`,
    where kernel function :math:`K_n` represents the Gaussian density.

    The unkown quantity :math:`\hat{\sigma}^2`
    can be estimated by the normalized weighted residual sum of squares:

    .. math::
        \hat{\sigma}^2(x_0) = \frac{\sum\limits_{i=1}^n
        (Y_i - \hat{Y}_i)^2 K_{h^*} (X_i - x_0)}
        { \text{tr} \{ \textbf{W}^* - \textbf{W}^* \textbf{X}^*
        ( \textbf{X}^{*T} \textbf{W}^* \textbf{X}^*)^{-1}
        \textbf{X}^{*T} \textbf{W}^*\} }.

    after fitting locally a `p`-th order polynomial for given
    ``x_0`` and pilot bandwidth :math:`h^*`. Analogously, the :math:`*`
    in the denominator indicate that this particular :math:`h^*` has been
    used to derive design matrix :math:`\textbf{X^*}` and
    weight matrix :math:`\textbf{W^*}`
    (:math:`\textbf{X}` and :math:`\textbf{W}` above).


    Arguments:
        x: Array of x data, predictor variable
        y: Array of y data, response variable
        degree: Degree of the polynomial.
        input_arr: Array or List with two entries
            0) ``x_0``: Evaluation point
            1) ``bw``: bandwidth

    Returns:
        Residual Squares Criterion for given
            combination of ``x_0`` and ``bw``.

    """
    # Unpack x0 and bandwidth
    x0, bw = input_arr

    delta = x - x0
    X = np.column_stack(
        [delta ** p for p in range(degree + 1)]
    )  # design matrix starting with np.ones column

    # Gaussian Kernel:  sigma = 1, mu = 0
    # https://en.wikipedia.org/wiki/Gaussian_function
    normpdf = np.exp(-((delta / bw) ** 2) / 2) * 1 / np.sqrt(2 * np.pi)
    weights = np.diag(normpdf)  # kernel weights

    # Wrap matrices
    S = X.T @ weights @ X
    S_inv = np.linalg.inv(S)

    cond_var_weights = S_inv @ (X.T @ (weights ** 2) @ X) @ S_inv
    V = cond_var_weights[
        0, 0
    ]  # first diagonal element reflects the effective number of local data points

    # Solution to weighted least-squares problem, see :cite:`Fan1996`, p.59 (3.5)
    beta_hat = S_inv @ X.T @ weights @ y
    y_hat = X @ beta_hat

    nominator = np.sum(((y - y_hat) ** 2) * weights)
    denominator = np.trace(weights - weights @ X @ S_inv @ X.T @ weights)
    weighted_res_sum_squares = nominator / denominator

    rsc = weighted_res_sum_squares * (1 + (degree + 1) * V)

    return rsc
